# -*- coding: utf-8 -*-
import os
from dataclasses import dataclass
from typing import List
from lxml.html import fromstring, tostring
from copy import deepcopy

@dataclass
class Component:
    el: str
    dom_el:str
    template:str
    style:str
    python:str

    def render(self, include_prefix:bool = True) -> str:
        if self.template.strip() == "":
            return ""

        if include_prefix:
            real_python = "# -*- coding: utf-8 -*-\nfrom mapsi import Mapsi, Component, DataStore\n\n" + self.python
        else:
            real_python = "\n" + self.python

        if "scoped" in self.style.split("\n")[0]:
            style_split = self.style.replace(self.style.split("\n")[0], "<style>").split("\n")
            style_scoped = []
            for line in style_split:
                if line.strip().endswith("{"):
                    if line.strip().startswith(self.dom_el):
                        style_scoped.append(line.replace(self.dom_el, f"{self.dom_el}>div"))
                    elif line.strip().startswith(self.el):
                        style_scoped.append(line.replace(self.el, f"{self.dom_el}>div"))
                    else:
                        style_scoped.append(line.replace(line.strip(), f"{self.dom_el}>div>{line.strip()}"))
                else:
                    style_scoped.append(line)

            real_style = "\n".join(style_scoped)
        else:
            real_style = self.style.replace(self.el, self.dom_el)

        class_line = f"class {self.el}(Component):"
        return real_python.replace(
            class_line,
            class_line + f'''
    template = """
{self.template}
    {real_style}"""
'''
        )

    def render_to_file(self, dest:str, mode:str = "w"):
        with open(dest, mode, encoding = "utf-8") as rtf:
            rtf.write(self.render(mode == "w"))


value_attrs = [ "model" ]
class Parser:
    @staticmethod
    def load_mapsi(mapsi_file:str) -> Component:
        with open(mapsi_file, "r", encoding = "utf-8") as mr:
            blocks = Parser.get_blocks(mr.read(), os.path.splitext(os.path.basename(mapsi_file))[0])

        return Component(**blocks)

    @staticmethod
    def load_mapsi_from_directory(root:str) -> List[Component]:
        from glob import glob

        return [
            Parser.load_mapsi(mapsi_file)
            for mapsi_file in glob(os.path.join(root, "*.mapsi"))
        ]

    @staticmethod
    def get_blocks(source:str, name:str) -> dict:
        blocks = { "template": "", "style": "", "el": name }
        if name.lower() == "app":
            blocks["dom_el"] = "main-app"
        else:
            blocks["dom_el"] = f"{name.lower()}-app"

        if "<template>" in source and "</template>" in source:
            blocks["template"] = source.split("<template>\n")[1].split("\n</template>")[0]
            blocks["template"] = Parser.render_template_as_dom(blocks["template"])

        if "<style" in source and "</style>" in source:
            style = source.split("<style")[1].split("\n</style>")[0]
            blocks["style"] = "<style" + style + "\n</style>"

        if "<python>" in source and "</python>" in source:
            python = source.split("<python>\n")[1].split("\n</python>")[0]
            blocks["python"] = "\n".join([
                line
                for line in python.split("\n")
                if not "pass" in line
            ])

            class_line = blocks["python"].split("\n")[0]
            class_line_to_change = f"""
class {name}(Component):
    def __init__(self):
        super().__init__()
"""

            if not name.lower() == "app":
                class_line_to_change = f'@Mapsi.register("{name.lower()}-app")' + class_line_to_change
            
            blocks["python"] = blocks["python"].replace(
                class_line, class_line_to_change
            # ).replace(
            #     "Component.", f"{name}."
            # ).replace(
            #     "self.", f"{name}."
            # ).replace(
            #     f"@{name}.", "@Component."
            ) + "\n"
        else:
            blocks["python"] = f"""
class {name}(Component):
    def __init__(self):
        super().__init__()
"""

            if not name.lower() == "app":
                blocks["python"] = f'@Mapsi.register("{name.lower()}-app")' + blocks["python"]

        return blocks

    @staticmethod
    def render_template_as_dom(template:str) -> str:
        if template == "":
            return template

        dom = fromstring(template)
        dom = Parser.change_dom_for_template(dom)
        return tostring(dom).decode("utf-8")

    @staticmethod
    def change_dom_for_template(dom):
        if not dom.text in ("", None):
            if "{" in dom.text and "}" in dom.text:
                dom.text = dom.text.replace("{ ", "{").replace(" }", "}")

        attribs = deepcopy(dom.attrib)
        for key, value in attribs.items():
            if key.startswith(":") and key[1:] in value_attrs:
                if not value.startswith("{"):
                    value = "{" + value
                
                if not value.startswith("}"):
                    value = value + "}"

                dom.attrib[key] = value

        for child in dom.getchildren():
            Parser.change_dom_for_template(child)

        return dom

    @staticmethod
    def change_template_for_format(template:str) -> str:
        return template.strip().replace("{ ", "{").replace(" }", "}")



# -*- coding: utf-8 -*-
from browser import document, webcomponent, html
from browser.template import Template


## routers
class RouterLink:
    def __init__(self):
        self.__name = self.getAttribute("to")
        self.removeAttribute("to")

        inner = self.innerHTML
        self.clear()

        container = html.DIV()
        container.innerHTML = inner
        container.bind("click", self.show_route)
        
        self.appendChild(container)

    def show_route(self, event):
        if Mapsi.router == None:
            raise RuntimeError("router-link can be used router-view exists!")

        document.select("router-view")[0].show_route(self.__name)

class RouterView:
    def show_route(self, name:str):
        self.innerHTML = ""

        for route in Mapsi.router:
            if route["url"] == name:
                self.innerHTML += f"<{route['el']} />"
            else:
                self.innerHTML += f"<!-- <{route['el']} /> -->"

    def render(self):
        self.innerHTML = ""
        for idx, route in enumerate(Mapsi.router):
            if idx == 0:
                self.innerHTML += f"<{route['name'].lower()}-app />"
            else:
                self.innerHTML += f"<!-- <{route['name'].lower()}-app /> -->"

class Router:
    def __init__(self, routes:list = []):
        self.routes = routes

webcomponent.define("router-link", RouterLink)
webcomponent.define("router-view", RouterView)


## DOM/Component
event_attrs = [ "click", "doubleClick", "change", "input" ]
value_attrs = [ "model" ]
value_attrs_to_change = {
    "model": "value"
}

EventStore:dict = {}
DataStore:dict = {}

class Eventer:
    def __init__(self, comp, event_name:str):
        self.__comp = comp
        self.__method = comp.methods[event_name]

    def call(self, event):
        self.__method(self.__comp, event)
        self.__comp.render_dom(
            self.__comp.firstChild,
            self.__comp.__class__.parameter(),
            self.__comp.__class__.data(),
            self.__comp.methods,
            False
        )

class VDOM:
    def __init__(self):
        self.dom_templates = {}

    def set_dom_templates(self, root):
        if not root == self.firstChild and root.childElementCount == 0:
            self.dom_templates[root] = root.innerText

        for child in root.childNodes:
            if not hasattr(child, "tagName"):
                continue

            self.set_dom_templates(child)

    def render_dom(self, dom, params, data, methods, bind_events:bool = True):
        format_info = {}
        if hasattr(params, "__iter__"):
            for key in params:
                format_info[key] = str(self.getAttribute(key))
                self.removeAttribute(key)
        
        if hasattr(data, "__iter__"):
            for key in data:
                format_info[key] = str(getattr(self, key))

        for child in dom.childNodes:
            if not hasattr(child, "tagName"):
                continue

            for key in child.attrs:
                value = child.attrs[key]

                if key.startswith(":"):
                    dom_key = key[1:]
                    if bind_events and dom_key in event_attrs:
                        child.bind(dom_key, Eventer(self, value).call)
                    elif dom_key in value_attrs:
                        child.setAttribute(value_attrs_to_change[dom_key], value.format(**format_info))

                    child.removeAttribute(key)

            if child.childElementCount == 0:
                child.innerText = self.dom_templates[child].format(**format_info)

            self.render_dom(child, data, methods, bind_events)

    # methods on registration
    def created(self):
        pass

    def mounted(self):
        pass

class Component(VDOM):
    template:str = ""
    el:str = ""

    def __init__(self):
        super().__init__()

        self.clear()
        # self.setAttribute("class", "mapsi-comp")

        self.innerHTML = self.replace_comp_tags()
        if "id" in self.firstChild.attrs:
            self.firstChild.removeAttribute("id")

        self.__dom_texts = {}
        self.set_dom_templates(self.firstChild)

        for name in self.__class__.parameter():
            setattr(self.__class__, name, self.__class__.parameter()[name])

        for name in self.__class__.data():
            setattr(self.__class__, name, self.__class__.data()[name])

        self.created()
        self.mounted()
        self.render()

    def replace_comp_tags(self) -> str:
        template = self.template.strip()
        for comp in Mapsi.components:
            if not comp in ("#comp-main-app", self.el):
                comp_tag = comp[6:-4].lower()
                dom_comp_tag = comp[6:]

                template = template.replace(
                    f"<{comp_tag} ", f"<{dom_comp_tag} "
                ).replace(
                    f"</{comp_tag}>", f"</{dom_comp_tag}>"
                )

        return template

    def render(self):
        params, data = self.__class__.parameter(), self.__class__.data()
        self.render_dom(self.firstChild, params, data, self.methods)

    @staticmethod
    def data() -> dict:
        return {}

    @staticmethod
    def parameter() -> dict:
        return {}

    @property
    def methods(self) -> dict:
        if self.el in EventStore:
            return EventStore[self.el]
        else:
            return {}

    @staticmethod
    def method(fn):
        def decorator(fn):
            comp_name = "#comp-" + str(fn).split(".")[0][10:].lower() + "-app"
            if not comp_name in EventStore:
                EventStore[comp_name] = {}

            EventStore[comp_name][fn.__name__] = fn
            return fn

        return decorator(fn)


## Mapsi
MapsiCompStyle = html.STYLE()
MapsiCompStyle.clear()
MapsiCompStyle.setAttribute("class", "mapsi-comp")
document.head <= MapsiCompStyle

class Mapsi:
    components:list = []
    router:list = []

    @staticmethod
    def use(obj):
        if isinstance(obj, Router):
            Mapsi.router = obj.routes

    @staticmethod
    def render(main_app:Component):
        Mapsi.register_component("main-app", main_app)

    @staticmethod
    def register(name:str):
        def decorator(comp:Component):
            Mapsi.register_component(name, comp)
            return comp

        return decorator

    @staticmethod
    def register_component(name:str, comp:Component):
        comp_name = f"#comp-{name}"
        setattr(comp, "el", comp_name)

        MapsiCompStyle.innerText += name + "{display: block;}"
        # print(123456789, document.select("style")[0].innerText)

        if not comp_name in Mapsi.components:
            Mapsi.components.append(comp_name)

        webcomponent.define(name, comp)

        if not comp.__name__ == "App":
            for idx, route in enumerate(Mapsi.router):
                if route["name"].lower() == comp.__name__.lower():
                    Mapsi.router[idx]["el"] = name



# -*- coding: utf-8 -*-
import os, argparse, shutil
from . import __path__ as mapsi_path


template_requires = ["public", "src", "App.mapsi", "main.py"]
def init_mapsi(dest:str):
    # create dest if not exists
    if not os.path.exists(dest):
        os.mkdir(dest)

    # check if project already exists
    if set(template_requires).issubset(os.listdir(dest)):
        raise RuntimeError("project files already exist!")

    # copy template files
    template_source = os.path.join(mapsi_path[0], "template_files")
    for t_file in os.listdir(template_source):
        if t_file in ("__pycache__"):
            continue

        t_path = os.path.join(template_source, t_file)
        if os.path.isdir(t_path):
            shutil.copytree(t_path, os.path.join(dest, t_file))
        else:
            shutil.copyfile(t_path, os.path.join(dest, t_file))

def build_mapsi(dest:str, make_dist:bool):
    from .core import Parser
    from glob import glob
    from brython import __path__ as brython_path

    source = os.getcwd()

    # check files exist
    if not set(template_requires).issubset(os.listdir(source)):
        raise RuntimeError("build can use in mapsi project directory!")

    # copy public files
    source_public_dir = os.path.join(source, "public")
    for public_file in os.listdir(source_public_dir):
        shutil.copyfile(os.path.join(source_public_dir, public_file), os.path.join(dest, public_file))

    # copy build files
    for build_file in os.listdir(os.path.join(mapsi_path[0], "build_files")):
        build_source_file = os.path.join(mapsi_path[0], "build_files", build_file)
        if os.path.isdir(build_source_file):
            shutil.copytree(build_source_file, os.path.join(dest, build_file))
        else:
            shutil.copyfile(build_source_file, os.path.join(dest, build_file))

    # copy brython files
    for script in ("brython.js", "brython_stdlib.js"):
        shutil.copyfile(os.path.join(brython_path[0], "data", script), os.path.join(dest, script))

    # copy project files
    shutil.copyfile(os.path.join(source, "main.py"), os.path.join(dest, "app_main.py"))
    if os.path.exists(os.path.join(source, "src", "router.py")):# router
        shutil.copyfile(os.path.join(source, "src", "router.py"), os.path.join(dest, "router.py"))

    if os.path.exists(os.path.join(source, "src", "assets")):# assets
        shutil.copytree(os.path.join(source, "src", "assets"), os.path.join(dest, "assets"))

    # load mapsi and compile to py
    Parser.load_mapsi(os.path.join(source, "App.mapsi")).render_to_file(os.path.join(dest, "App.py"))
    dest_views_file, dest_components_file = os.path.join(dest, "views.py"), os.path.join(dest, "components.py")
    with open(dest_views_file, "w", encoding = "utf-8") as dvfw:
        dvfw.write("")

        for idx, comp in enumerate(Parser.load_mapsi_from_directory(os.path.join(source, "src", "views"))):
            dvfw.write(comp.render(idx == 0))

    with open(dest_components_file, "w", encoding = "utf-8") as dcfw:
        dcfw.write("")

        for idx, comp in enumerate(Parser.load_mapsi_from_directory(os.path.join(source, "src", "components"))):
            dcfw.write(comp.render(idx == 0))

    if make_dist:# build modules by brython-cli
        os.chdir(dest)

        # build
        os.system("brython-cli --modules")

        # remove python files
        for py_file in glob(os.path.join(dest, "*.py"), recursive = False):
            os.remove(py_file)

        # remove unnecessary files
        os.remove(os.path.join(dest, "brython_stdlib.js"))

        os.chdir(source)

def serve_mapsi(cwd:str):
    import sys

    cur_cwd = os.getcwd()
    os.chdir(cwd)
    os.system(f"{sys.executable} -m http.server --bind=127.0.0.1")
    os.chdir(cur_cwd)


def main():
    parser = argparse.ArgumentParser(description = "Mapsi CLI")
    parser.add_argument("job", type = str, help = "job to execute(init, serve, dist)")
    parser.add_argument("--dest", type = str, help = "destination for job", default = ".")
    args = vars(parser.parse_args())

    job, dest = args["job"], args["dest"]
    if dest.startswith("./") or dest.startswith(".\\"):
        dest = dest[2:]
    elif dest.startswith("."):
        dest = dest[1:]

    if job == "init":
        if not os.path.isabs(dest):
            dest = os.path.join(os.getcwd(), dest)

        init_mapsi(dest)

    elif job in ("build", "serve"):
        if not os.path.isabs(dest):
            dest = os.path.join(os.getcwd(), "dist" if dest == "" else dest)

        if os.path.exists(dest):
            shutil.rmtree(dest)

        os.mkdir(dest)

        if job == "build":
            build_mapsi(dest, True)
        else:
            build_mapsi(dest, False)
            serve_mapsi(dest)
            
if __name__ == "__main__":
    import sys
    sys.exit(main())



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup
from mapsi import __author__, __email__, __version__

with open("requirements.txt", "r", encoding = "utf-8") as require_read:
    requires = require_read.readlines()

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

setup(
    author=__author__,
    author_email=__email__,
    version=__version__,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Korean",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    name="mapsipy",
    description="MVVM UI Framework for Python and Brython",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    license="MIT license",
    keywords=["mapsipy", "brython", "mvvm"],
    entry_points={
        "console_scripts": [
            "mapsipy-cli=mapsi.cli:main",
        ],
    },
    include_package_data=False,
    install_requires=requires,
    setup_requires=requires,
    packages=["mapsi", "mapsi/assets", "mapsi/build_files", "mapsi/template_files"],
    package_data={
        "": [
            "**/*.*"
        ]
    },
    url="https://github.com/eseunghwan/mapsipy",
    zip_safe=False
)



# -*- coding: utf-8 -*-
from browser import document, html
from mapsi import Mapsi

def setup():
    from App import App as main_app
    import app_main

    real_app_elm = html.maketag("main-app")()
    document.body.removeChild(document.select("App")[0])
    document.body <= real_app_elm

    app_main.setup()
    Mapsi.render(main_app)

    import views
    import components

    may_router_view = document.select("router-view")
    if len(may_router_view) > 0:
        may_router_view[0].render()



# -*- coding: utf-8 -*-

routes = [
    {
        "name": "Home",
        "url": "Home"
    },
    {
        "name": "Counter",
        "url": "Counter"
    },
    {
        "name": "Hello",
        "url": "Hello"
    }
]
